#!/usr/bin/env python3
#
# ===- clang-tidy-diff.py - ClangTidy Diff Checker -----------*- python -*--===#
#
# Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
# See https://llvm.org/LICENSE.txt for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# ===-----------------------------------------------------------------------===#

r"""
ClangTidy Diff Checker
======================

This script reads input from a unified diff, runs clang-tidy on all changed
files and outputs clang-tidy warnings in changed lines only. This is useful to
detect clang-tidy regressions in the lines touched by a specific patch.
Example usage for git/svn users:

  git diff -U0 HEAD^ | clang-tidy-diff.py -p1
  svn diff --diff-cmd=diff -x-U0 | \
      clang-tidy-diff.py -fix -checks=-*,modernize-use-override

"""

import argparse
import glob
import json
import multiprocessing
import os
import re
import shutil
import subprocess  #nosec The usage of subprocess was reviewed, and system shell will not be invoked.
import sys
import tempfile
import threading
import traceback

try:
    import yaml
except ImportError:
    yaml = None

is_py2 = sys.version[0] == "2"

if is_py2:
    import Queue as queue
else:
    import queue as queue


def run_tidy(task_queue, lock, timeout, failed_files):
    watchdog = None
    while True:
        command = task_queue.get()
        try:
            proc = subprocess.Popen(
                command, stdout=subprocess.PIPE, stderr=subprocess.PIPE
            )

            if timeout is not None:
                watchdog = threading.Timer(timeout, proc.kill)
                watchdog.start()

            stdout, stderr = proc.communicate()
            if proc.returncode != 0:
                if proc.returncode < 0:
                    msg = "Terminated by signal %d : %s\n" % (
                        -proc.returncode,
                        " ".join(command),
                    )
                    stderr += msg.encode("utf-8")
                failed_files.append(command)

            with lock:
                sys.stdout.write(stdout.decode("utf-8") + "\n")
                sys.stdout.flush()
                if stderr:
                    sys.stderr.write(stderr.decode("utf-8") + "\n")
                    sys.stderr.flush()
        except Exception as e:
            with lock:
                sys.stderr.write("Failed: " + str(e) + ": ".join(command) + "\n")
        finally:
            with lock:
                if not (timeout is None or watchdog is None):
                    if not watchdog.is_alive():
                        sys.stderr.write(
                            "Terminated by timeout: " + " ".join(command) + "\n"
                        )
                    watchdog.cancel()
            task_queue.task_done()


def start_workers(max_tasks, tidy_caller, arguments):
    for _ in range(max_tasks):
        t = threading.Thread(target=tidy_caller, args=arguments)
        t.daemon = True
        t.start()


def merge_replacement_files(tmpdir, mergefile):
    """Merge all replacement files in a directory into a single file"""
    # The fixes suggested by clang-tidy >= 4.0.0 are given under
    # the top level key 'Diagnostics' in the output yaml files
    mergekey = "Diagnostics"
    merged = []
    for replacefile in glob.iglob(os.path.join(tmpdir, "*.yaml")):
        content = yaml.safe_load(open(replacefile, "r"))
        if not content:
            continue  # Skip empty files.
        merged.extend(content.get(mergekey, []))

    if merged:
        # MainSourceFile: The key is required by the definition inside
        # include/clang/Tooling/ReplacementsYaml.h, but the value
        # is actually never used inside clang-apply-replacements,
        # so we set it to '' here.
        output = {"MainSourceFile": "", mergekey: merged}
        with open(mergefile, "w") as out:
            yaml.safe_dump(output, out)
    else:
        # Empty the file:
        open(mergefile, "w").close()


def main():
    parser = argparse.ArgumentParser(
        description="Run clang-tidy against changed files, and "
        "output diagnostics only for modified "
        "lines."
    )
    parser.add_argument(
        "-clang-tidy-binary",
        metavar="PATH",
        default="clang-tidy",
        help="path to clang-tidy binary",
    )
    parser.add_argument(
        "-p",
        metavar="NUM",
        default=0,
        help="strip the smallest prefix containing P slashes",
    )
    parser.add_argument(
        "-regex",
        metavar="PATTERN",
        default=None,
        help="custom pattern selecting file paths to check "
        "(case sensitive, overrides -iregex)",
    )
    parser.add_argument(
        "-iregex",
        metavar="PATTERN",
        default=r".*\.(cpp|cc|c\+\+|cxx|c|cl|h|hpp|m|mm|inc)",
        help="custom pattern selecting file paths to check "
        "(case insensitive, overridden by -regex)",
    )
    parser.add_argument(
        "-j",
        type=int,
        default=1,
        help="number of tidy instances to be run in parallel.",
    )
    parser.add_argument(
        "-timeout", type=int, default=None, help="timeout per each file in seconds."
    )
    parser.add_argument(
        "-fix", action="store_true", default=False, help="apply suggested fixes"
    )
    parser.add_argument(
        "-checks",
        help="checks filter, when not specified, use clang-tidy " "default",
        default="",
    )
    parser.add_argument(
        "-config-file",
        dest="config_file",
        help="Specify the path of .clang-tidy or custom config file",
        default="",
    )
    parser.add_argument("-use-color", action="store_true", help="Use colors in output")
    parser.add_argument(
        "-path", dest="build_path", help="Path used to read a compile command database."
    )
    if yaml:
        parser.add_argument(
            "-export-fixes",
            metavar="FILE_OR_DIRECTORY",
            dest="export_fixes",
            help="A directory or a yaml file to store suggested fixes in, "
            "which can be applied with clang-apply-replacements. If the "
            "parameter is a directory, the fixes of each compilation unit are "
            "stored in individual yaml files in the directory.",
        )
    else:
        parser.add_argument(
            "-export-fixes",
            metavar="DIRECTORY",
            dest="export_fixes",
            help="A directory to store suggested fixes in, which can be applied "
            "with clang-apply-replacements. The fixes of each compilation unit are "
            "stored in individual yaml files in the directory.",
        )
    parser.add_argument(
        "-extra-arg",
        dest="extra_arg",
        action="append",
        default=[],
        help="Additional argument to append to the compiler " "command line.",
    )
    parser.add_argument(
        "-extra-arg-before",
        dest="extra_arg_before",
        action="append",
        default=[],
        help="Additional argument to prepend to the compiler " "command line.",
    )
    parser.add_argument(
        "-quiet",
        action="store_true",
        default=False,
        help="Run clang-tidy in quiet mode",
    )
    parser.add_argument(
        "-load",
        dest="plugins",
        action="append",
        default=[],
        help="Load the specified plugin in clang-tidy.",
    )
    parser.add_argument(
        "-allow-no-checks",
        action="store_true",
        help="Allow empty enabled checks.",
    )

    clang_tidy_args = []
    argv = sys.argv[1:]
    if "--" in argv:
        clang_tidy_args.extend(argv[argv.index("--") :])
        argv = argv[: argv.index("--")]

    args = parser.parse_args(argv)

    # Extract changed lines for each file.
    filename = None
    lines_by_file = {}
    for line in sys.stdin:
        match = re.search('^\\+\\+\\+\\ "?(.*?/){%s}([^ \t\n"]*)' % args.p, line)
        if match:
            filename = match.group(2)
        if filename is None:
            continue

        if args.regex is not None:
            if not re.match("^%s$" % args.regex, filename):
                continue
        else:
            if not re.match("^%s$" % args.iregex, filename, re.IGNORECASE):
                continue

        match = re.search(r"^@@.*\+(\d+)(,(\d+))?", line)
        if match:
            start_line = int(match.group(1))
            line_count = 1
            if match.group(3):
                line_count = int(match.group(3))
            if line_count == 0:
                continue
            end_line = start_line + line_count - 1
            lines_by_file.setdefault(filename, []).append([start_line, end_line])

    if not any(lines_by_file):
        print("No relevant changes found.")
        sys.exit(0)

    max_task_count = args.j
    if max_task_count == 0:
        max_task_count = multiprocessing.cpu_count()
    max_task_count = min(len(lines_by_file), max_task_count)

    combine_fixes = False
    export_fixes_dir = None
    delete_fixes_dir = False
    if args.export_fixes is not None:
        # if a directory is given, create it if it does not exist
        if args.export_fixes.endswith(os.path.sep) and not os.path.isdir(
            args.export_fixes
        ):
            os.makedirs(args.export_fixes)

        if not os.path.isdir(args.export_fixes):
            if not yaml:
                raise RuntimeError(
                    "Cannot combine fixes in one yaml file. Either install PyYAML or specify an output directory."
                )

            combine_fixes = True

        if os.path.isdir(args.export_fixes):
            export_fixes_dir = args.export_fixes

    if combine_fixes:
        export_fixes_dir = tempfile.mkdtemp()
        delete_fixes_dir = True

    # Tasks for clang-tidy.
    task_queue = queue.Queue(max_task_count)
    # A lock for console output.
    lock = threading.Lock()

    # List of files with a non-zero return code.
    failed_files = []

    # Run a pool of clang-tidy workers.
    start_workers(
        max_task_count, run_tidy, (task_queue, lock, args.timeout, failed_files)
    )

    # Form the common args list.
    common_clang_tidy_args = []
    if args.fix:
        common_clang_tidy_args.append("-fix")
    if args.checks != "":
        common_clang_tidy_args.append("-checks=" + args.checks)
    if args.config_file != "":
        common_clang_tidy_args.append("-config-file=" + args.config_file)
    if args.quiet:
        common_clang_tidy_args.append("-quiet")
    if args.build_path is not None:
        common_clang_tidy_args.append("-p=%s" % args.build_path)
    if args.use_color:
        common_clang_tidy_args.append("--use-color")
    if args.allow_no_checks:
        common_clang_tidy_args.append("--allow-no-checks")
    for arg in args.extra_arg:
        common_clang_tidy_args.append("-extra-arg=%s" % arg)
    for arg in args.extra_arg_before:
        common_clang_tidy_args.append("-extra-arg-before=%s" % arg)
    for plugin in args.plugins:
        common_clang_tidy_args.append("-load=%s" % plugin)

    for name in lines_by_file:
        line_filter_json = json.dumps(
            [{"name": name, "lines": lines_by_file[name]}], separators=(",", ":")
        )

        # Run clang-tidy on files containing changes.
        command = [args.clang_tidy_binary]
        command.append("-line-filter=" + line_filter_json)
        if args.export_fixes is not None:
            # Get a temporary file. We immediately close the handle so clang-tidy can
            # overwrite it.
            (handle, tmp_name) = tempfile.mkstemp(suffix=".yaml", dir=export_fixes_dir)
            os.close(handle)
            command.append("-export-fixes=" + tmp_name)
        command.extend(common_clang_tidy_args)
        command.append(name)
        command.extend(clang_tidy_args)

        task_queue.put(command)

    # Application return code
    return_code = 0

    # Wait for all threads to be done.
    task_queue.join()
    # Application return code
    return_code = 0
    if failed_files:
        return_code = 1

    if combine_fixes:
        print("Writing fixes to " + args.export_fixes + " ...")
        try:
            merge_replacement_files(export_fixes_dir, args.export_fixes)
        except:
            sys.stderr.write("Error exporting fixes.\n")
            traceback.print_exc()
            return_code = 1

    if delete_fixes_dir:
        shutil.rmtree(export_fixes_dir)
    sys.exit(return_code)


if __name__ == "__main__":
    main()



# This file is a minimal clang-format sublime-integration. To install:
# - Change 'binary' if clang-format is not on the path (see below).
# - Put this file into your sublime Packages directory, e.g. on Linux:
#     ~/.config/sublime-text-2/Packages/User/clang-format-sublime.py
# - Add a key binding:
#     { "keys": ["ctrl+shift+c"], "command": "clang_format" },
#
# With this integration you can press the bound key and clang-format will
# format the current lines and selections for all cursor positions. The lines
# or regions are extended to the next bigger syntactic entities.
#
# It operates on the current, potentially unsaved buffer and does not create
# or save any files. To revert a formatting, just undo.

from __future__ import absolute_import, division, print_function
import sublime
import sublime_plugin
import subprocess  #nosec The usage of subprocess was reviewed, and system shell will not be invoked.

# Change this to the full path if clang-format is not on the path.
binary = "clang-format"

# Change this to format according to other formatting styles. See the output of
# 'clang-format --help' for a list of supported styles. The default looks for
# a '.clang-format' or '_clang-format' file to indicate the style that should be
# used.
style = None


class ClangFormatCommand(sublime_plugin.TextCommand):
    def run(self, edit):
        encoding = self.view.encoding()
        if encoding == "Undefined":
            encoding = "utf-8"
        regions = []
        command = [binary]
        if style:
            command.extend(["-style", style])
        for region in self.view.sel():
            regions.append(region)
            region_offset = min(region.a, region.b)
            region_length = abs(region.b - region.a)
            command.extend(
                [
                    "-offset",
                    str(region_offset),
                    "-length",
                    str(region_length),
                    "-assume-filename",
                    str(self.view.file_name()),
                ]
            )
        old_viewport_position = self.view.viewport_position()
        buf = self.view.substr(sublime.Region(0, self.view.size()))
        p = subprocess.Popen(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            stdin=subprocess.PIPE,
        )
        output, error = p.communicate(buf.encode(encoding))
        if error:
            print(error)
        self.view.replace(
            edit, sublime.Region(0, self.view.size()), output.decode(encoding)
        )
        self.view.sel().clear()
        for region in regions:
            self.view.sel().add(region)
        # FIXME: Without the 10ms delay, the viewport sometimes jumps.
        sublime.set_timeout(
            lambda: self.view.set_viewport_position(old_viewport_position, False), 10
        )



"""
Minimal clang-rename integration with Vim.

Before installing make sure one of the following is satisfied:

* clang-rename is in your PATH
* `g:clang_rename_path` in ~/.vimrc points to valid clang-rename executable
* `binary` in clang-rename.py points to valid to clang-rename executable

To install, simply put this into your ~/.vimrc for python2 support

    noremap <leader>cr :pyf <path-to>/clang-rename.py<cr>

For python3 use the following command (note the change from :pyf to :py3f)

    noremap <leader>cr :py3f <path-to>/clang-rename.py<cr>

IMPORTANT NOTE: Before running the tool, make sure you saved the file.

All you have to do now is to place a cursor on a variable/function/class which
you would like to rename and press '<leader>cr'. You will be prompted for a new
name if the cursor points to a valid symbol.
"""

from __future__ import absolute_import, division, print_function
import vim
import subprocess  #nosec The usage of subprocess was reviewed, and system shell will not be invoked.
import sys


def main():
    binary = "clang-rename"
    if vim.eval('exists("g:clang_rename_path")') == "1":
        binary = vim.eval("g:clang_rename_path")

    # Get arguments for clang-rename binary.
    offset = int(vim.eval('line2byte(line("."))+col(".")')) - 2
    if offset < 0:
        print(
            "Couldn't determine cursor position. Is your file empty?", file=sys.stderr
        )
        return
    filename = vim.current.buffer.name

    new_name_request_message = "type new name:"
    new_name = vim.eval("input('{}\n')".format(new_name_request_message))

    # Call clang-rename.
    command = [
        binary,
        filename,
        "-i",
        "-offset",
        str(offset),
        "-new-name",
        str(new_name),
    ]
    # FIXME: make it possible to run the tool on unsaved file.
    p = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = p.communicate()

    if stderr:
        print(stderr)

    # Reload all buffers in Vim.
    vim.command("checktime")


if __name__ == "__main__":
    main()



# This file is a minimal clang-include-fixer vim-integration. To install:
# - Change 'binary' if clang-include-fixer is not on the path (see below).
# - Add to your .vimrc:
#
#   noremap <leader>cf :pyf path/to/llvm/source/tools/clang/tools/extra/clang-include-fixer/tool/clang-include-fixer.py<cr>
#
# This enables clang-include-fixer for NORMAL and VISUAL mode. Change
# "<leader>cf" to another binding if you need clang-include-fixer on a
# different key.
#
# To set up clang-include-fixer, see
# http://clang.llvm.org/extra/clang-include-fixer.html
#
# With this integration you can press the bound key and clang-include-fixer will
# be run on the current buffer.
#
# It operates on the current, potentially unsaved buffer and does not create
# or save any files. To revert a fix, just undo.

from __future__ import print_function
import argparse
import difflib
import json
import re
import subprocess  #nosec The usage of subprocess was reviewed, and system shell will not be invoked.
import vim

# set g:clang_include_fixer_path to the path to clang-include-fixer if it is not
# on the path.
# Change this to the full path if clang-include-fixer is not on the path.
binary = "clang-include-fixer"
if vim.eval('exists("g:clang_include_fixer_path")') == "1":
    binary = vim.eval("g:clang_include_fixer_path")

maximum_suggested_headers = 3
if vim.eval('exists("g:clang_include_fixer_maximum_suggested_headers")') == "1":
    maximum_suggested_headers = max(
        1, vim.eval("g:clang_include_fixer_maximum_suggested_headers")
    )

increment_num = 5
if vim.eval('exists("g:clang_include_fixer_increment_num")') == "1":
    increment_num = max(1, vim.eval("g:clang_include_fixer_increment_num"))

jump_to_include = False
if vim.eval('exists("g:clang_include_fixer_jump_to_include")') == "1":
    jump_to_include = vim.eval("g:clang_include_fixer_jump_to_include") != "0"

query_mode = False
if vim.eval('exists("g:clang_include_fixer_query_mode")') == "1":
    query_mode = vim.eval("g:clang_include_fixer_query_mode") != "0"


def GetUserSelection(message, headers, maximum_suggested_headers):
    eval_message = message + "\n"
    for idx, header in enumerate(headers[0:maximum_suggested_headers]):
        eval_message += "({0}). {1}\n".format(idx + 1, header)
    eval_message += "Enter (q) to quit;"
    if maximum_suggested_headers < len(headers):
        eval_message += " (m) to show {0} more candidates.".format(
            min(increment_num, len(headers) - maximum_suggested_headers)
        )

    eval_message += "\nSelect (default 1): "
    res = vim.eval("input('{0}')".format(eval_message))
    if res == "":
        # choose the top ranked header by default
        idx = 1
    elif res == "q":
        raise Exception("   Insertion cancelled...")
    elif res == "m":
        return GetUserSelection(
            message, headers, maximum_suggested_headers + increment_num
        )
    else:
        try:
            idx = int(res)
            if idx <= 0 or idx > len(headers):
                raise Exception()
        except Exception:
            # Show a new prompt on invalid option instead of aborting so that users
            # don't need to wait for another clang-include-fixer run.
            print("Invalid option: {}".format(res), file=sys.stderr)
            return GetUserSelection(message, headers, maximum_suggested_headers)
    return headers[idx - 1]


def execute(command, text):
    # Avoid flashing a cmd prompt on Windows.
    startupinfo = None
    if sys.platform.startswith("win32"):
        startupinfo = subprocess.STARTUPINFO()
        startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
        startupinfo.wShowWindow = subprocess.SW_HIDE

    p = subprocess.Popen(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        stdin=subprocess.PIPE,
        startupinfo=startupinfo,
    )
    return p.communicate(input=text.encode("utf-8"))


def InsertHeaderToVimBuffer(header, text):
    command = [
        binary,
        "-stdin",
        "-insert-header=" + json.dumps(header),
        vim.current.buffer.name,
    ]
    stdout, stderr = execute(command, text)
    if stderr:
        raise Exception(stderr)
    if stdout:
        lines = stdout.splitlines()
        sequence = difflib.SequenceMatcher(None, vim.current.buffer, lines)
        line_num = None
        for op in reversed(sequence.get_opcodes()):
            if op[0] != "equal":
                vim.current.buffer[op[1] : op[2]] = lines[op[3] : op[4]]
            if op[0] == "insert":
                # line_num in vim is 1-based.
                line_num = op[1] + 1

        if jump_to_include and line_num:
            vim.current.window.cursor = (line_num, 0)


# The vim internal implementation (expand("cword"/"cWORD")) doesn't support
# our use case very well, we re-implement our own one.
def get_symbol_under_cursor():
    line = vim.eval('line(".")')
    # column number in vim is 1-based.
    col = int(vim.eval('col(".")')) - 1
    line_text = vim.eval("getline({0})".format(line))
    if len(line_text) == 0:
        return ""
    symbol_pos_begin = col
    p = re.compile("[a-zA-Z0-9:_]")
    while symbol_pos_begin >= 0 and p.match(line_text[symbol_pos_begin]):
        symbol_pos_begin -= 1

    symbol_pos_end = col
    while symbol_pos_end < len(line_text) and p.match(line_text[symbol_pos_end]):
        symbol_pos_end += 1
    return line_text[symbol_pos_begin + 1 : symbol_pos_end]


def main():
    parser = argparse.ArgumentParser(
        description="Vim integration for clang-include-fixer"
    )
    parser.add_argument("-db", default="yaml", help="clang-include-fixer input format.")
    parser.add_argument("-input", default="", help="String to initialize the database.")
    # Don't throw exception when parsing unknown arguments to make the script
    # work in neovim.
    # Neovim (at least v0.2.1) somehow mangles the sys.argv in a weird way: it
    # will pass additional arguments (e.g. "-c script_host.py") to sys.argv,
    # which makes the script fail.
    args, _ = parser.parse_known_args()

    # Get the current text.
    buf = vim.current.buffer
    text = "\n".join(buf)

    if query_mode:
        symbol = get_symbol_under_cursor()
        if len(symbol) == 0:
            print("Skip querying empty symbol.")
            return
        command = [
            binary,
            "-stdin",
            "-query-symbol=" + get_symbol_under_cursor(),
            "-db=" + args.db,
            "-input=" + args.input,
            vim.current.buffer.name,
        ]
    else:
        # Run command to get all headers.
        command = [
            binary,
            "-stdin",
            "-output-headers",
            "-db=" + args.db,
            "-input=" + args.input,
            vim.current.buffer.name,
        ]
    stdout, stderr = execute(command, text)
    if stderr:
        print(
            "Error while running clang-include-fixer: {}".format(stderr),
            file=sys.stderr,
        )
        return

    include_fixer_context = json.loads(stdout)
    query_symbol_infos = include_fixer_context["QuerySymbolInfos"]
    if not query_symbol_infos:
        print("The file is fine, no need to add a header.")
        return
    symbol = query_symbol_infos[0]["RawIdentifier"]
    # The header_infos is already sorted by clang-include-fixer.
    header_infos = include_fixer_context["HeaderInfos"]
    # Deduplicate headers while keeping the order, so that the same header would
    # not be suggested twice.
    unique_headers = []
    seen = set()
    for header_info in header_infos:
        header = header_info["Header"]
        if header not in seen:
            seen.add(header)
            unique_headers.append(header)

    if not unique_headers:
        print("Couldn't find a header for {0}.".format(symbol))
        return

    try:
        selected = unique_headers[0]
        inserted_header_infos = header_infos
        if len(unique_headers) > 1:
            selected = GetUserSelection(
                "choose a header file for {0}.".format(symbol),
                unique_headers,
                maximum_suggested_headers,
            )
            inserted_header_infos = [
                header for header in header_infos if header["Header"] == selected
            ]
        include_fixer_context["HeaderInfos"] = inserted_header_infos

        InsertHeaderToVimBuffer(include_fixer_context, text)
        print("Added #include {0} for {1}.".format(selected, symbol))
    except Exception as error:
        print(error, file=sys.stderr)
    return


if __name__ == "__main__":
    main()
